#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

"""
Project-wide **string munging** (i.e., generic low-level operations
transforming strings into new derivative strings) utilities.

This private submodule is *not* intended for importation by downstream callers.
"""

# ....................{ IMPORTS                            }....................
from beartype.roar._roarwarn import _BeartypeUtilCallableWarning
# from beartype.typing import Dict
from beartype._cave._cavefast import NumberType
from beartype._data.typing.datatyping import TypeWarning
from beartype._data.kind.datakindtext import CHARS_PUNCTUATION
from beartype._util.utilobject import get_object_basename_scoped_or_none
from collections.abc import Callable
from pprint import saferepr

# ....................{ REPRESENTERS                       }....................
def represent_object(
    # Mandatory parameters.
    obj: object,

    # Optional parameters.
    max_len: int = 96,
) -> str:
    '''
    Pretty-printed quasi-human-readable variant of the string returned by the
    non-pretty-printed machine-readable :meth:`obj.__repr__` dunder method of
    the passed object, truncated to the passed maximum string length.

    Specifically, this function (in order):

    #. Obtains this object's representation by calling ``repr(object)``.
    #. If this representation is neither suffixed by a punctuation character
       (i.e., character in the standard :attr:`string.punctuation` set) *nor*
       representing a byte-string whose representations are prefixed by ``b'``
       and suffixed by ``'`` (e.g., ``b'Check, mate.'``), double-quotes this
       representation for disambiguity with preceding characters -- notably,
       sequence indices. Since strings returned by this function commonly
       follow sequence indices in error messages, failing to disambiguate the
       two produces non-human-readable output:

           >>> def wat(mate: typing.List[str]) -> int: return len(mate)
           >>> get_func_pith_violation(
           ...     func=muh_func, pith_name='mate', pith_value=[7,])
           beartype.roar.BeartypeCallHintParamViolation: @beartyped wat()
           parameter mate=[7] violates PEP type hint typing.List[str], as list
           item 0 value 7 not a str.

       Note the substring "item 0 value 7", which misreads like a blatant bug.
       Double-quoting the "7" suffices to demarcate values from indices.
    #. If this representation exceeds the passed maximum length, replaces the
       suffix of this representation exceeding this length with an ellipses
       (i.e., ``"..."`` substring).

    Caveats
    -------
    **This function is unavoidably slow and should thus not be called from
    optimized performance-critical code.** This function internally performs
    mildly expensive operations, including iterating-based string munging.
    Ideally, this function should *only* be called to create user-oriented
    exception messages where performance is a negligible concern.

    **This function preserves all quote-protected newline characters** (i.e.,
    ``"\\n"``) **in this representation.** Since the :meth:`str.__repr__`
    dunder method implicitly quote-protects all newlines in the original
    string, this function effectively preserves all newlines in strings.

    Parameters
    ----------
    obj : object
        Object to be represented.
    max_len: int, optional
        Maximum length of the string to be returned. Defaults to a standard
        line length of 100 characters minus output indentation of 4 characters.

    Returns
    -------
    str
        Pretty-printed quasi-human-readable variant of this object's
        non-pretty-printed machine-readable representation.
    '''
    assert isinstance(max_len, int), f'{repr(max_len)} not integer.'

    #FIXME: [SPEED] Note that the popular BSD-licensed Celerity project's
    #celerity.utils.saferepr.saferepr() function claims to actually be faster
    #than the repr() builtin under certain circumstances. While impressive,
    #repurposing Celerity's saferepr() implementation for @beartype will be
    #non-trivial; that function internally leverages a number of non-trivial
    #internal functions, including a streaming iterator that appears to be
    #performing some sort of ad-hoc tokenization (!) on the input object's
    #string representation. Although that submodule is less than 300 lines,
    #that's 300 *INTENSE* lines. Until someone pounds their fists on our issue
    #tracker (so, never), let's pretend no one cares. See also:
    #   https://github.com/celery/celery/blob/master/celery/utils/saferepr.py

    # For efficiency, initially attempt to *NON-RECURSIVELY* introspect the
    # machine-readable string describing this object. Note that:
    # * This representation quote-protects all newlines in this representation.
    #   Ergo, "\n" *MUST* be matched as r"\n" instead below.
    # * For debuggability, the verbose (albeit less readable) output of repr()
    #   is preferred to the terse (albeit more readable) output of str().
    # * For safety, the pprint.saferepr() function explicitly protected against
    #   recursive data structures *WOULD* typically be preferred to the unsafe
    #   repr() builtin *NOT* protected against such recursion. Sadly,
    #   pprint.saferepr() is extremely unoptimized and thus susceptible to
    #   extreme performance regressions when passed a worst-case object (e.g.,
    #   deeply nested container).
    try:
        obj_repr = repr(obj)
    # If doing so fails with a recursion error, the passed object is recursive.
    # In this case, fallback to recursively introspecting the machine-readable
    # string describing this object. Note that doing so is *EXTREMELY*
    # inefficient and thus performed only as a last-ditch desperate fallback.
    except RecursionError:
        obj_repr = saferepr(obj)

    # If this representation is empty, return empty double-quotes. Although
    # most objects (including outlier singletons like "None" and the empty
    # string) have non-empty representations, caller-defined classes may
    # maliciously override the __repr__() dunder method to return an empty
    # string rather than the representation of an empty string (i.e., '""').
    if not obj_repr:
        return '""'
    # Else, this representation is non-empty.
    #
    # If this representation is neither...
    elif not (
        # Prefixed by punctuation *NOR*...
        obj_repr[0] in CHARS_PUNCTUATION or
        # An instance of a class whose representations do *NOT* benefit from
        # explicit quoting...
        isinstance(obj, _TYPES_UNQUOTABLE)
    ):
    # Then this representation is *NOT* demarcated from preceding characters in
    # the parent string embedding this representation. In this case,
    # double-quote this representation for disambiguity with preceding
    # characters (e.g., sequence indices).
        obj_repr = f'"{obj_repr}"'
    # Else, this representation is either prefixed by punctuation *OR* an
    # instance of a class whose representations do *NOT* benefit from quoting.

    # If this representation exceeds this maximum length...
    if len(obj_repr) > max_len:
        # Avoid circular import dependencies.
        from beartype._util.text.utiltextmunge import truncate_str

        # Truncate this representation to this maximum length.
        obj_repr = truncate_str(text=obj_repr, max_len=max_len)
        # print(f'obj repr truncated: {obj_repr}')
    # Else, this representation is less than or equal to this maximum length.

    # Return this representation.
    return obj_repr

# ....................{ REPRESENTER ~ callable             }....................
def represent_func(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    warning_cls: TypeWarning = _BeartypeUtilCallableWarning,
) -> str:
    '''
    Machine-readable representation of the passed callable.

    Caveats
    -------
    **This function is unavoidably slow and should thus not be called from
    optimized performance-critical code.** This function internally performs
    extremely expensive operations, including abstract syntax tree (AST)-based
    parsing of Python scripts and modules deserialized from disk. Ideally, this
    function should *only* be called to create user-oriented exception messages
    where performance is a negligible concern.

    Parameters
    ----------
    func : Callable
        Callable to be represented.
    warning_cls : TypeWarning, optional
        Type of warning to be emitted in the event of a non-fatal error.
        Defaults to :class:`_BeartypeUtilCallableWarning`.

    Warns
    -----
    :class:`warning_cls`
        If this callable is a pure-Python lambda function whose definition is
        *not* parsable from the script or module defining that lambda.

    Returns
    -------
    str
        Machine-readable representation of that callable.
    '''
    assert callable(func), f'{repr(func)} not callable.'

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunccode import get_func_code_or_none
    from beartype._util.func.utilfunctest import is_func_lambda

    # If that callable is a pure-Python lambda function, return either:
    # * If this lambda is defined by an on-disk script or module source file,
    #   the exact substring of that file defining this lambda.
    # * Else (e.g., if this lambda is dynamically defined in-memory), a
    #   placeholder string.
    if is_func_lambda(func):
        return (
            get_func_code_or_none(func=func, warning_cls=warning_cls) or
            '<lambda>'
        )
    # Else, that callable is *NOT* a pure-Python lambda function.

    #FIXME: Actually, we should be calling a new get_object_name_or_none()
    #function instead -- but that function currently doesn't exist and we're
    #lazy. The issue with get_object_basename_scoped_or_none() is that this
    #getter fails to return the module name of this function. *shrug*
    func_basename_scoped = get_object_basename_scoped_or_none(func)

    # If that callable is named, return this name.
    if func_basename_scoped:
        return func_basename_scoped
    # Else, that callable is unnamed due to failing to define both the
    # "__qualname__" and "__name__" dunder attributes and thus have *NO* names.
    # Although most callables are named, some are not. This includes:
    # * Callable "functools.partial" objects.

    # Return the machine-readable representation of that callable as a fallback.
    return repr(func)

# ....................{ REPRESENTERS ~ pith                }....................
def represent_pith(
    # Mandatory parameters.
    pith: object,

    # Optional parameters.
    is_color: bool = True,
) -> str:
    '''
    Human-readable description of the passed **pith** (i.e., arbitrary object
    violating the current type check) intended to be embedded in an exception
    message explaining this violation.

    Parameters
    ----------
    pith : object
        Arbitrary object violating the current type check.
    is_color : bool, optional
        :data:`True` only if embellishing this label with colour. Defaults to
        :data:`True` for convenience.

    Returns
    -------
    str
        Human-readable description of this object.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextlabel import label_pith_value
    from beartype._util.text.utiltextprefix import prefix_pith_type

    # Create and return this representation.
    return (
        f'{prefix_pith_type(pith=pith, is_color=is_color)}'
        f'{label_pith_value(pith=pith, is_color=is_color)}'
    )

# ....................{ PRIVATE ~ globals                  }....................
_TYPES_UNQUOTABLE = (
    # Byte strings, whose representations are already quoted as "b'...'".
    bytes,
    # Numbers, whose representations are guaranteed to both contain *NO*
    # whitespace and be sufficiently terse as to benefit from *NO* quoting.
    NumberType,
)
'''
**Unquotable tuple union** (i.e., isinstancable tuple of all classes such that
the :func:`represent_object` function intentionally avoids double-quoting the
machine-readable representations all instances of these classes, typically due
to these representations either being effectively quoted already *or*
sufficiently terse as to not benefit from being quoted).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`484`- and :pep:`585`-compliant unit tests.

This submodule unit tests :pep:`484` and :pep:`585` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ sync                       }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Blargh!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
def test_decor_sync_generator() -> None:
    '''
    Test decorating synchronous generators with the :func:`beartype.beartype`
    decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintParamViolation,
        BeartypeDecorHintPep484585Exception,
    )
    from beartype_test._util.pytroar import raises_uncached
    from collections.abc import (
        Generator as Pep585Generator,
        Iterable as Pep585Iterable,
        Iterator as Pep585Iterator,
    )
    from inspect import isgeneratorfunction
    from typing import (
        Union,
        Generator as Pep484Generator,
        Iterable as Pep484Iterable,
        Iterator as Pep484Iterator,
    )

    # ....................{ CALLABLES                      }....................
    @beartype
    def transcendent_joy(
        said_the: Union[str, float], big_green_bat: Union[str, int]):
        '''
        :func:`beartype.beartype`-decorated synchronous generator whose return
        is intentionally left unannotated, validating a common edge case.
        '''

        # Yield an arbitrary combination of the passed parameters.
        yield said_the + big_green_bat

    # ....................{ LOCALS                         }....................
    # Tuple of all PEP 484- or 585-compliant type hints annotating the returns
    # of synchronous generators covering all possible edge cases.
    HINTS_RETURN = (
        # ....................{ NON-PEP                    }....................
        # The unsubscripted "collections.abc.Generator" ABC is a
        # PEP-noncompliant type.
        Pep585Generator,

        # ....................{ PEP 484                    }....................
        Pep484Generator[Union[str, float], None, None],
        Pep484Iterable[Union[str, float]],
        Pep484Iterator[Union[str, float]],

        # ....................{ PEP 585                    }....................
        Pep585Generator[Union[str, float], None, None],
        Pep585Iterable[Union[str, float]],
        Pep585Iterator[Union[str, float]],
    )

    # ....................{ PASS                           }....................
    # For each return type hint defined above...
    for hint_return in HINTS_RETURN:
        @beartype
        def western_logocentric_stuff(
            said_the: Union[str, int], bigger_greener_bat: Union[str, float]
        ) -> hint_return:
            '''
            :func:`beartype.beartype`-decorated synchronous generator whose
            return is annotated by a :pep:`484`- or :pep:`585`-compliant type
            hint deeply type-checking the value yielded by this generator.
            '''

            # Yield an arbitrary combination of the passed parameters.
            yield said_the + bigger_greener_bat

        # Assert that the high-level wrapper generated by the @beartype
        # decorator type-checking the low-level synchronous generator defined
        # above preserves "inspect"-based metadata describing this generator.
        # Specifically, assert that this wrapper is still a generator.
        assert isgeneratorfunction(western_logocentric_stuff) is True

        # Assert that iterating the synchronous generator created by passing
        # this function valid parameters yields the expected item.
        assert next(western_logocentric_stuff(
              'all my Western logocentric stuff ', 'about factoring numbers',
        )) == 'all my Western logocentric stuff about factoring numbers'

        # Assert that iterating the synchronous generator created by passing
        # this function invalid parameters raises the expected violation.
        with raises_uncached(BeartypeCallHintParamViolation):
            next(western_logocentric_stuff(
                b'asking the entities you meet', b'to factor large numbers'))

    # Assert that the high-level wrapper generated by the @beartype decorator
    # type-checking the low-level synchronous generator defined above preserves
    # "inspect"-based metadata describing this generator.
    assert isgeneratorfunction(transcendent_joy) is True

    # Assert that iterating the synchronous generator created by passing this
    # function valid parameters yields the expected item.
    assert next(transcendent_joy(
        'First one of wine, then one of brine, ',
        'then one more yet of turpentine,',
    )) == (
        'First one of wine, then one of brine, '
        'then one more yet of turpentine,'
    )

    # ....................{ FAIL                           }....................
    # Assert that iterating the synchronous generator created by passing this
    # function invalid parameters raises the expected violation.
    with raises_uncached(BeartypeCallHintParamViolation):
        next(transcendent_joy(
            b'Not facts or factors or factories',
            b'but contact with the abstract attractor that brings you back to me'
        ))

    # Assert that @beartype raises the expected exception when decorating a
    # synchronous generator annotating its return as anything *EXCEPT*
    # "Generator[...]", "Iterable[...]", and "Iterator[...]".
    with raises_uncached(BeartypeDecorHintPep484585Exception):
        @beartype
        def GET_OUT_OF_THE_CAR(
            FOR_THE_LOVE_OF_GOD: str, FACTOR_THE_NUMBER: str) -> str:
            # Return an arbitrary object (sorta) satisfying this return hint.
            yield FOR_THE_LOVE_OF_GOD + FACTOR_THE_NUMBER

# ....................{ TESTS ~ async                      }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Urgh!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
async def test_decor_async_coroutine() -> None:
    '''
    Test decorating **coroutines** (i.e., asynchronous non-generator functions)
    with the :func:`beartype.beartype` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from asyncio import sleep
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintParamViolation,
        BeartypeDecorHintPep585Exception,
    )
    from beartype_test._util.pytroar import raises_uncached
    from collections.abc import Coroutine as Pep585Coroutine
    from inspect import (
        isasyncgenfunction,
        iscoroutinefunction,
    )
    from typing import (
        Union,
        Coroutine as Pep484Coroutine,
    )

    # ....................{ LOCALS                         }....................
    # Tuple of all PEP 484- or 585-compliant type hints annotating the returns
    # of synchronous generators covering all possible edge cases.
    HINTS_RETURN = (
        # ....................{ PEP 484                    }....................
        Pep484Coroutine[None, None, Union[str, float]],
        Union[str, float],

        # ....................{ PEP 585                    }....................
        Pep585Coroutine[None, None, Union[str, float]],
    )

    # ....................{ PASS                           }....................
    # For each return type hint defined above...
    for hint_return in HINTS_RETURN:
        @beartype
        async def control_the_car(
            said_the: Union[str, int], biggest_greenest_bat: Union[str, float],
        ) -> hint_return:
            '''
            :func:`beartype.beartype`-decorated coroutine whose return is
            annotated by a :pep:`484`- or :pep:`585`-compliant type hint deeply
            type-checking the value returned by this coroutine.
            '''

            # Silently reduce to an asynchronous noop. *ALL* coroutines are
            # required to call the "await" keyword at least once. Since the
            # placeholder object returned below is synchronous and thus *CANNOT*
            # be asynchronously awaited, this fallback wrapper has *NO* recourse
            # but to asynchronously await a minimal-cost awaitable. Aaaaaaand...
            #
            # This is why the "asyncio" API is Python's most hated. We sigh!
            await sleep(0)

            # Return an arbitrary combination of the passed parameters.
            return said_the + biggest_greenest_bat

        # Assert that the high-level wrapper generated by the @beartype
        # decorator type-checking the low-level coroutine defined above
        # preserves "inspect"-based metadata describing this coroutine.
        # Specifically, assert that this wrapper is still a coroutine but has
        # *NOT* been improperly transformed into an asynchronous generator.
        assert iscoroutinefunction(control_the_car) is True
        assert isasyncgenfunction(control_the_car) is False

        # Assert that awaiting the coroutine created by passing this function
        # valid parameters returns the expected value.
        assert await control_the_car(
            'I saw the big green bat bat a green big eye. ',
            'Suddenly I knew I had gone too far.') == (
            'I saw the big green bat bat a green big eye. '
            'Suddenly I knew I had gone too far.')

        # Assert that awaiting the coroutine created by passing this function
        # invalid parameters raises the expected violation.
        with raises_uncached(BeartypeCallHintParamViolation):
            await control_the_car(
                b'The sea was made of strontium;',
                b'the beach was made of rye.',
            )

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating a
    # coroutine whose return is annotated by a PEP 585-compliant coroutine type
    # hint *NOT* subscripted by exactly three child hints.
    with raises_uncached(BeartypeDecorHintPep585Exception):
        @beartype
        async def with_each_planck_moment_ever_fit() -> (
            Pep585Coroutine['to be eternally enjoyed']):
            # Silently reduce to an asynchronous noop. See above for details.
            await sleep(0)

            # Return an arbitrary object (sorta) satisfying this return hint.
            return 'Time will decay us but time can be left blank'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook managers** (i.e., lower-level private-facing functions
internally driving the higher-level public facing import hooks exported by the
:mod:`beartype.claw._clawmain` submodule).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.claw._package.clawpkgenum import BeartypeClawCoverage
from beartype.claw._package.clawpkgtrie import (
    PackagesTrieBlacklist,
    PackagesTrieBlacklisted,
    PackagesTrieWhitelist,
    # iter_packages_trie,
    # remove_beartype_pathhook_unless_packages_trie,
)
from beartype.claw._package._clawpkgmake import (
    make_conf_hookable,
    make_package_names_from_args,
)
from beartype.claw._importlib.clawimpmain import (
    add_beartype_pathhook,
    # remove_beartype_pathhook,
)
from beartype.roar import BeartypeClawHookException
from beartype.typing import Optional
from beartype._conf.confmain import BeartypeConf
from beartype._data.typing.datatyping import IterableStrs
from beartype._util.py.utilpyinterpreter import is_python_optimized
from collections.abc import (
    Iterable as IterableABC,
)

# ....................{ HOOKERS                            }....................
#FIXME: Unit test us up, please.
def hook_packages(
    # Keyword-only arguments.
    *,

    # Mandatory keyword-only arguments.
    claw_coverage: BeartypeClawCoverage,
    conf: BeartypeConf,

    # Optional keyword-only arguments.
    package_name: Optional[str] = None,
    package_names: Optional[IterableStrs] = None,
) -> None:
    '''
    Register a new **beartype package import path hook** (i.e., callable
    inserted to the front of the standard :mod:`sys.path_hooks` list recursively
    applying the :func:`beartype.beartype` decorator to all typed callables and
    classes of all submodules of all packages with the passed names on the first
    importation of those submodules).

    Parameters
    ----------
    claw_coverage : BeartypeClawCoverage
        **Import hook coverage** (i.e., competing package scope over which to
        apply the path hook added by this function, each with concomitant
        tradeoffs with respect to runtime complexity and quality assurance).
    conf : BeartypeConf
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
    package_name : Optional[str]
        Either:

        * If ``coverage`` is :attr:`.BeartypeClawCoverage.PACKAGES_ONE`, the
          fully-qualified name of the package to be type-checked.
        * Else, ignored.

        Defaults to :data:`None`.
    package_names : Optional[Iterable[str]]]
        Either:

        * If ``coverage`` is :attr:`.BeartypeClawCoverage.PACKAGES_MANY`, an
          iterable of the fully-qualified names of one or more packages to be
          type-checked.
        * Else, ignored.

        Defaults to :data:`None`.

    Raises
    ------
    BeartypeClawHookException
        If either:

        * The passed ``package_names`` parameter is either:

          * Neither a string nor an iterable (i.e., fails to satisfy the
            :class:`collections.abc.Iterable` protocol).
          * An empty string or iterable.
          * A non-empty string that is *not* a valid **package name** (i.e.,
            ``"."``-delimited concatenation of valid Python identifiers).
          * A non-empty iterable containing at least one item that is either:

            * *Not* a string.
            * The empty string.
            * A non-empty string that is *not* a valid **package name** (i.e.,
              ``"."``-delimited concatenation of valid Python identifiers).

        * The passed ``conf`` parameter is *not* a beartype configuration (i.e.,
          :class:`.BeartypeConf` instance).

    See Also
    --------
    https://stackoverflow.com/a/43573798/2809027
        StackOverflow answer strongly inspiring the low-level implementation of
        this function with respect to inscrutable :mod:`importlib` machinery.
    '''
    # print(f'[hook_packages]:')

    # ....................{ PREAMBLE                       }....................
    # If the active Python interpreter is optimized either at process-invocation
    # time (e.g., by the user passing one or more "-O" command-line options *OR*
    # setting the '${PYTHONOPTIMIZE}" environment variable to a positive integer
    # when the active Python interpreter was forked) *OR* after
    # process-invocation time (e.g., by the user setting the '${PYTHONOPTIMIZE}"
    # environment variable to a positive integer in an interactive REPL). Our
    # awesome userbase requested this.
    if is_python_optimized():
        return 

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_lock

    # ....................{ LOCALS                         }....................
    # Replace this beartype configuration (which is typically unsuitable for
    # usage in import hooks) with a new beartype configuration suitable for
    # usage in import hooks.
    conf = make_conf_hookable(conf)

    # Iterable of the passed fully-qualified names of all packages to be hooked.
    package_names = make_package_names_from_args(
        claw_coverage=claw_coverage,
        conf=conf,
        package_name=package_name,
        package_names=package_names,
    )

    # ....................{ HOOKS                          }....................
    # With a submodule-specific thread-safe reentrant lock...
    with claw_lock:
        # ....................{ BLACKLIST                  }....................
        # If blacklisting one or more packages from type-checking, do so.
        # print(f'Blacklisting packages: {repr(conf.claw_skip_package_names)}')
        if conf.claw_skip_package_names:
            _blacklist_packages(conf.claw_skip_package_names)
        # Else, *NO* packages are being blacklisted from type-checking. Fine!

        # ....................{ WHITELIST ~ beartype_all   }....................
        # If type-checking *ALL* packages, do so.
        if claw_coverage is BeartypeClawCoverage.PACKAGES_ALL:
            _whitelist_packages_all(conf)
        # ....................{ WHITELIST ~ beartype_packa }....................
        # Else, only a subset of packages are being type-checked. Do it! Do it!
        else:
            _whitelist_packages_some(package_names=package_names, conf=conf)  # type: ignore[arg-type]

        # ....................{ path hook                  }....................
        # Lastly, if our beartype import path hook singleton has *NOT* already
        # been added to the standard "sys.path_hooks" list, do so now.
        #
        # Note that we intentionally:
        # * Do so in a thread-safe manner *INSIDE* this lock.
        # * Defer doing so until *AFTER* the above iteration has successfully
        #   registered the desired packages with our global trie. Why? This path
        #   hook subsequently calls the companion get_package_conf_or_none()
        #   function, which accesses this trie.
        add_beartype_pathhook()

# ....................{ PRIVATE ~ blacklisters             }....................
#FIXME: Docstring us up, please.
def _blacklist_packages(package_names: IterableStrs) -> None:
    '''
    Recursively **blacklist** (i.e., prevent import hooks from implicitly
    runtime type-checking) any subsequently imported packages and modules whose
    absolute names are either directly listed in the passed iterable *or* which
    are subpackages or submodules transitively residing in any packages or
    modules whose absolute names are directly listed in this iterable.

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.

    Parameters
    ----------
    package_names : IterableStrs
        Iterable of the absolute names of all packages to be blacklisted.
    '''
    assert isinstance(package_names, IterableABC), (
        f'{repr(package_names)} not iterable.')

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # For the fully-qualified name of each package to be blacklisted...
    for package_name in package_names:
        # List of each unqualified basename comprising this name, split from
        # this fully-qualified name on "." delimiters.
        #
        # Note that the "str.split('.')" and "str.rsplit('.')" calls produce the
        # same lists under all edge cases. We arbitrarily call the former rather
        # than the latter for simplicity.
        package_basenames = package_name.split('.')

        # Trailing unqualified basename of the desired target
        # (sub)package to be specifically blacklisted (e.g.,
        # "some_subpackage" given the list "['some_package',
        # 'some_subpackage']").
        #
        # Note that this index is guaranteed to be safe. See the
        # following discussion for further commentary.
        package_basename_last = package_basenames[-1]

        # Ignore the trailing unqualified basename of the desired target
        # (sub)package to be specifically blacklisted (e.g., reduce the
        # list "['some_package', 'some_subpackage']" to merely the list
        # "['some_package']"). Why? Because the loop performed below
        # only finds the direct parent package of the desired target
        # (sub)package to be specifically blacklisted. That target
        # (sub)package itself is *NOT* looped over.
        #
        # Note that this slice is guaranteed to be safe. By prior
        # validation performed at "BeartypeConf" instantiation time,
        # this iterable is guaranteed to contain *ONLY* non-empty
        # strings. Then "skip_package_basenames" is guaranteed to be a
        # non-empty list of one or more unqualified basenames. After
        # performing this slice, "skip_package_basenames" then reduces
        # to a possibly -empty list of zero or more such basenames.
        package_basenames = package_basenames[:-1]

        # Current subtrie of the global trie blacklist describing the
        # currently iterated basename of this package, initialized to
        # the global trie blacklist blacklisting all top-level packages.
        subpackages_trie_blacklist = claw_state.packages_trie_blacklist
        # print(f'Blacklist: {repr(claw_state.packages_trie_blacklist)}')

        # For each unqualified basename comprising the directed path
        # from the root package of this package through all intermediary
        # parent packages of this package down to this package...
        for package_basename in package_basenames:
            # If this parent package has yet to be described by a prior
            # call to this function, add a new subtrie blacklist
            # describing this parent package.
            if package_basename not in subpackages_trie_blacklist:
                subpackages_trie_blacklist[package_basename] = (
                    PackagesTrieBlacklist(package_basename=package_basename))
            # Else, this parent package has already been described by a
            # prior call to this function.
            #
            # In either case, a subtrie whitelist describing this parent
            # package is now guaranteed to exist.
            # print(f'Visiting blacklisting parent subpackage "{package_basename}"...')

            # Iterate the current subtrie one subpackage deeper.
            subpackages_trie_blacklist = subpackages_trie_blacklist[
                package_basename]
        # The "skip_package_basenames" list contains at least one
        # basename. The above iteration thus set the currently visited
        # "subpackages_trie_blacklist" to at least one subtrie of the
        # global trie blacklist. More importantly, this subtrie is
        # guaranteed to describe the direct parent package of the
        # current (sub)package being blacklisted.
        # print(f'Blacklisted package "{package_name}" trie {repr(subpackages_trie_blacklist)}...')
        # print(f'Blacklisting leaf subpackage "{package_basename_last}"...')

        # Blacklist this (sub)package from its direct parent package.
        #
        # Note that this seemingly trivial assignment adroitly handles
        # three distinct cases:
        # * If the subtrie
        #   "subpackages_trie_blacklist[skip_package_basename]" does
        #   *NOT* already exist, this assignment behaves as expected.
        # * Else, the subtrie
        #   "subpackages_trie_blacklist[skip_package_basename]" already
        #   exists. In this case, this subtrie is either:
        #   * "PackagesTrieBlacklisted", implying this package has
        #     already been blacklisted by a prior call of this function.
        #     In this case, reblacklisting this package effectively
        #     reduces to an efficient noop.
        #   * *ANY* other "PackagesTrieBlacklist" subtrie, implying that
        #     a prior call of this function previously blacklisted one
        #     or more subpackages of this package (but *NOT* this
        #     package itself). In this case, blacklisting this package
        #     expands that previous blacklisting of one or more
        #     subpackages of this package to the entire package. Score!
        subpackages_trie_blacklist[package_basename_last] = (
            PackagesTrieBlacklisted)

# ....................{ PRIVATE ~ whitelisters             }....................
def _whitelist_packages_all(conf: BeartypeConf) -> None:
    '''
    Globally **whitelist** (i.e., implicitly runtime type-check) *all*
    subsequently imported packages and modules.

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.

    Parameters
    ----------
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
    '''
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # Beartype configuration currently associated with *ALL* packages by a prior
    # call to this function if any *OR* "None" (i.e., if this function has yet
    # to be called under this Python interpreter).
    conf_curr = claw_state.packages_trie_whitelist.conf_if_hooked

    # If the higher-level beartype_all() function (calling this lower-level
    # whitelister) has yet to be called under this interpreter, associate this
    # configuration with *ALL* packages.
    if conf_curr is None:
        claw_state.packages_trie_whitelist.conf_if_hooked = conf
    # Else, beartype_all() was already called under this interpreter.
    #
    # If the caller passed a different configuration to that prior call than
    # that passed to this current call, raise an exception.
    elif conf_curr != conf:
        raise BeartypeClawHookException(
            f'beartype_all() previously passed '
            f'conflicting beartype configuration:\n'
            f'\t----------( OLD "conf" PARAMETER )----------\n'
            f'\t{repr(conf_curr)}\n'
            f'\t----------( NEW "conf" PARAMETER )----------\n'
            f'\t{repr(conf)}\n'
        )
    # Else, the caller passed the same configuration to that prior call than
    # that passed to the current call. In this case, silently reduce to a noop.


def _whitelist_packages_some(
    package_names: IterableStrs,
    conf: BeartypeConf,
) -> None:
    '''
    Globally **whitelist** (i.e., implicitly runtime type-check) the proper
    subset of subsequently imported packages and modules whose absolute names
    are either directly listed in the passed iterable *or* which are subpackages
    or submodules transitively residing in any packages or modules whose
    absolute names are directly listed in this iterable.

    Caveats
    -------
    **This function is only safely callable in a thread-safe manner from within
    a** ``with claw_lock:`` **context manager.** Equivalently, this function is
    *not* safely callable outside that manager.

    Parameters
    ----------
    package_names : Iterable[str]
        Iterable of the absolute names of all packages to be whitelisted.
    conf : BeartypeConf
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
    '''
    assert isinstance(package_names, IterableABC), (
        f'{repr(package_names)} not iterable.')
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # For the fully-qualified name of each package to be whitelisted...
    for package_name in package_names:  # type: ignore[union-attr]
        # List of each unqualified basename comprising this name, split from
        # this fully-qualified name on "." delimiters.
        #
        # Note that the "str.split('.')" and "str.rsplit('.')" calls produce the
        # same lists under all edge cases. We arbitrarily call the former rather
        # than the latter for simplicity.
        package_basenames = package_name.split('.')

        # Current subtrie of the global trie whitelist describing the currently
        # iterated basename of this package, initialized to the global trie
        # whitelist configuring all top-level packages.
        subpackages_trie_whitelist = claw_state.packages_trie_whitelist

        # For each unqualified basename comprising the directed path from the
        # root package of this package through all intermediary parent packages
        # of this package down to this package...
        for package_basename in package_basenames:
            # If this parent package has yet to be described by a prior call to
            # this function, add a new subtrie whitelist describing this parent
            # package.
            if package_basename not in subpackages_trie_whitelist:
                subpackages_trie_whitelist[package_basename] = (
                    PackagesTrieWhitelist(package_basename=package_basename))
            # Else, this parent package has already been described by a prior
            # call to this function.
            #
            # In either case, a subtrie whitelist describing this parent package
            # is now guaranteed to exist.

            # Iterate the current subtrie one subpackage deeper.
            subpackages_trie_whitelist = subpackages_trie_whitelist[
                package_basename]
        # The "package_basenames" list contains at least one basename. The above
        # iteration thus set the currently visited "subpackages_trie_whitelist"
        # to at least one subtrie of the global trie whitelist. More
        # importantly, this subtrie is guaranteed to describe the current
        # (sub)package being whitelisted.
        # print(f'Hooked package "{package_name}" subpackage trie {repr(subpackages_trie)}...')

        # Beartype configuration currently associated with this package by a
        # prior call to this function if any *OR* "None" (i.e., if this package
        # has yet to be hooked by such a call).
        conf_curr = subpackages_trie_whitelist.conf_if_hooked

        # If this package has yet to be whitelisted by a prior call to this
        # function, associate this configuration with this package.
        if conf_curr is None:
            subpackages_trie_whitelist.conf_if_hooked = conf
        # Else, that package was already whitelisted by such a call.
        #
        # If the caller passed a different configuration to this current call
        # than that prior call, raise an exception.
        elif conf_curr != conf:
            raise BeartypeClawHookException(
                f'Beartype import hook '
                f'(e.g., beartype.claw.beartype_*() function) '
                f'previously passed conflicting beartype configuration for '
                f'package "{package_name}":\n'
                f'\t----------( OLD "conf" PARAMETER )----------\n'
                f'\t{repr(conf_curr)}\n'
                f'\t----------( NEW "conf" PARAMETER )----------\n'
                f'\t{repr(conf)}\n'
            )
        # Else, the caller passed the same configuration to both the prior and
        # current calls. In this case, ignore this redundant request to
        # rewhitelist this package.

# ....................{ UNHOOKERS                          }....................
#FIXME: Preserved in perpetuity. Who knows what was going on here? Unhooking is
#almost certainly useful -- but this is useless until (A) someone actually
#requests this functionality and (B) we actually test this extensively. *sigh*
#FIXME: [WAT] *UHM.* This is great, *BUT NEVER ACTUALLY EXPOSED TO USERS OR
#INTERNALLY CALLED ANYWHERE.* Seriously. What was the original game plan here?
#No idea. Never documented it. Consider:
#* For each permanent public "beartype.claw" import hook (e.g., beartype_all()),
#  define a corresponding temporary public "beartype.claw" import hook
#  internally deferring to this unhook_packages() function: e.g.,
#      @contextmanager
#      def beartyping_all(*args, **kwargs) -> None:
#          try:
#              beartype_all(*args, **kwargs)
#          finally:
#              #FIXME: Not quite right, obviously. We'll need to pass the proper
#              #parameters like "claw_coverage". Blah, blah. Boring! Next.
#              #FIXME: Uhm... wait. Should we even be calling unhook_packages()
#              #here? Probably not. We should probably just be snapshotting the
#              #existing "claw_state" or something. Shaking my head, fam.
#              unhook_packages(*args, **kwargs)
#FIXME: Unit test us up, please. Since this is never internally called anywhere,
#this is currently completely untested (and probably totally broken). Yikes.
# def unhook_packages(
#     # Keyword-only arguments.
#     *,
#
#     # Mandatory keyword-only arguments.
#     claw_coverage: BeartypeClawCoverage,
#     conf: BeartypeConf,
#
#     # Optional keyword-only arguments.
#     package_name: Optional[str] = None,
#     package_names: Optional[Iterable[str]] = None,
# ) -> None:
#     '''
#     Unregister a previously registered **beartype package import path hook**
#     (i.e., callable inserted to the front of the standard :mod:`sys.path_hooks`
#     list recursively applying the :func:`beartype.beartype` decorator to all
#     typed callables and classes of all submodules of all packages with the
#     passed names on the first importation of those submodules).
#
#     See Also
#     --------
#     :func:`.hook_packages`
#         Further details.
#     '''
#
#     # ....................{ IMPORTS                        }....................
#     # Avoid circular import dependencies.
#     from beartype.claw._clawstate import (
#         claw_lock,
#         claw_state,
#     )
#
#     # ....................{ LOCALS                         }....................
#     # Replace this beartype configuration (which is typically unsuitable for
#     # usage in import hooks) with a new beartype configuration suitable for
#     # usage in import hooks.
#     conf = make_conf_hookable(conf)
#
#     # Iterable of the passed fully-qualified names of all packages to be
#     # unhooked.
#     package_names = make_package_names_from_args(
#         claw_coverage=claw_coverage,
#         conf=conf,
#         package_name=package_name,
#         package_names=package_names,
#     )
#
#     # ....................{ WHITELIST                      }....................
#     # With a submodule-specific thread-safe reentrant lock...
#     with claw_lock:
#         # ....................{ beartype_all()             }....................
#         # If the caller requested all-packages coverage...
#         if claw_coverage is BeartypeClawCoverage.PACKAGES_ALL:
#             # Unhook the beartype configuration previously associated with *ALL*
#             # packages by a prior call to the beartype_all() function.
#             claw_state.packages_trie_whitelist.conf_if_hooked = None
#         # ....................{ beartype_packages()        }....................
#         # Else, the caller requested coverage over a subset of packages. In this
#         # case...
#         else:
#             # For the fully-qualified names of each package to be
#             # unregistered...
#             for package_name in package_names:  # type: ignore[union-attr]
#                 # List of all subpackages tries describing each parent package
#                 # transitively containing the passed package (as well as that of
#                 # that package itself).
#                 subpackages_tries = list(iter_packages_trie(package_name))
#
#                 # Reverse this list in-place, such that:
#                 # * The first item of this list is the subpackages trie
#                 #   describing that package itself.
#                 # * The last item of this list is the subpackages trie
#                 #   describing the root package of that package.
#                 subpackages_tries.reverse()
#
#                 # Unhook the beartype configuration previously associated with
#                 # that package by a prior call to the hook_packages() function.
#                 subpackages_tries[0].conf_if_hooked = None
#
#                 # Child sub-subpackages trie of the currently iterated
#                 # subpackages trie, describing the child subpackage of the
#                 # current parent package transitively containing that package.
#                 subsubpackages_trie = None
#
#                 # For each subpackages trie describing a parent package
#                 # transitively containing that package...
#                 for subpackages_trie in subpackages_tries:
#                     # If this is *NOT* the first iteration of this loop (in
#                     # which case this subpackages trie is a parent package
#                     # rather than that package itself) *AND*...
#                     if subsubpackages_trie is not None:
#                         # If this child sub-subpackages trie describing this
#                         # child sub-subpackage has one or more children, then
#                         # this child sub-subpackages trie still stores
#                         # meaningful metadata and is thus *NOT* safely
#                         # deletable. Moreover, this implies that:
#                         # * *ALL* parent subpackages tries of this child
#                         #   sub-subpackages trie also still store meaningful
#                         #   metadata and are thus also *NOT* safely deletable.
#                         # * There exists no more meaningful work to be performed
#                         #   by this iteration. Ergo, we immediately halt this
#                         #   iteration now.
#                         if subsubpackages_trie:
#                             break
#                         # Else, this child sub-subpackages trie describing this
#                         # child sub-subpackage has *NO* children, implying this
#                         # child sub-subpackages trie no longer stores any
#                         # meaningful metadata and is thus safely deletable.
#
#                         # Unqualified basename of this child sub-subpackage.
#                         subsubpackage_basename = (
#                             subsubpackages_trie.package_basename)
#
#                         # Delete this child sub-subpackages trie from this
#                         # parent subpackages trie.
#                         del subpackages_trie[subsubpackage_basename]  # pyright: ignore
#                     # Else, this is the first iteration of this loop.
#
#                     # Treat this parent subpackages trie as the child
#                     # sub-subpackages trie in the next iteration of this loop.
#                     subsubpackages_trie = subpackages_trie
#
#         # ....................{ path hook                  }....................
#         # Lastly, if *ALL* meaningful metadata has now been removed from our
#         # global trie, remove our beartype import path hook singleton from the
#         # standard "sys.path_hooks" list.
#         #
#         # Note that we intentionally:
#         # * Do so in a thread-safe manner *INSIDE* this lock.
#         # * Defer doing so until *AFTER* the above iteration has successfully
#         #   unregistered the desired packages with our global trie.
#         remove_beartype_pathhook_unless_packages_trie()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype sanified type hint metadata dataclass** (i.e., class aggregating
*all* metadata returned by :mod:`beartype._check.convert.convmain` functions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Optional,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.metadata.hint.hintsane import HintSane
from beartype._data.typing.datatypingport import Hint
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_hint_recursive(
    # Mandatory parameters.
    hint: Hint,
    hint_parent_sane: Optional[HintSane],

    # Optional parameters.
    hint_recursable_depth_max: int = 0,
) -> bool:
    '''
    :data:`True` only if the passed **recursable type hint** (i.e., type hint
    implicitly supporting recursion like, say, a :pep:`695`-compliant type
    alias) is actually **recursive** (i.e., has already been visited by the
    current breadth-first search (BFS) over all type hints transitively nested
    in some root type hint) with respect to the passed previously sanified
    metadata for the parent type hint of the passed type hint.

    Caveats
    -------
    **This tester assumes this hint to be hashable.** Although *most*
    PEP-compliant hints are hashable, some are not (e.g., :pep:`593`-compliant
    metahints annotated by unhashable objects like ``typing.Annotated[object,
    []]``). Callers that cannot guarantee this hint to be hashable should
    protect calls to this tester inside a ``try`` block explicitly catching the
    :exc:`TypeError` exception this tester raises when this hint is unhashable:

    .. code-block:: python

       # Attempt to test whether hint is recursive or not.
       try:
           if is_hint_recursive(hint=hint, hint_parent_sane=hint_parent_sane):
               pass
       # If doing so raises a "TypeError", this hint is unhashable and thus
       # inapplicable for hint recursion. In this case, ignore this hint.
       except TypeError:
           pass

    Parameters
    ----------
    hint : Hint
        Recursable type hint to be inspected.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If this recursable hint is a root type hint, :data:`None`.
        * Else, **sanified parent type hint metadata** (i.e., immutable and thus
          hashable object encapsulating *all* metadata previously returned by
          :mod:`beartype._check.convert.convmain` sanifiers after sanitizing
          the possibly PEP-noncompliant parent hint of this recursable hint into
          a fully PEP-compliant parent hint).
    hint_recursable_depth_max : int, default: 0
        **Maximum recursion depth** (i.e., maximum total number of times this
        recursable hint is permitted to have been previously visited during the
        current search from the root type hint down to this recursable hint
        before considering this recursable hint to be "recursive").
        Specifically, this tester returns either:

        * If the total number of times this recursable hint has been previously
          visited is less than or equal to this integer, :data:`False`.
        * If the total number of times this recursable hint has been previously
          visited is greater than this integer, :data:`True`.

        Defaults to 0, in which case this recursable hint is considered to be
        "recursive" if this hint has been visited at least once before.

    Returns
    -------
    bool
        :data:`True` only if this recursable type hint is recursive.

    Raises
    ------
    TypeError
        If this hint is unhashable.
    '''
    assert isinstance(hint_parent_sane, NoneTypeOr[HintSane]), (
        f'{repr(hint_parent_sane)} neither sanified hint metadata nor "None".')
    assert isinstance(hint_recursable_depth_max, int), (
        f'{repr(hint_recursable_depth_max)} not integer.')
    assert hint_recursable_depth_max >= 0, (
        f'{repr(hint_recursable_depth_max)} < 0.')

    # True only if...
    is_hint_recursive_state = (
        # This hint has a parent *AND*...
        hint_parent_sane is not None and
        # The total number of times this hint has been visited is greater than
        # this maximum depth, implying this hint to be a transitive parent of
        # itself a sufficient number of times to consider this hint recursive.
        (
            hint_parent_sane.hint_recursable_to_depth.get(hint, 0) >
            hint_recursable_depth_max
        )
    )
    # print(f'Hint {hint} with parent {hint_parent_sane} recursive? {is_hint_recursive_state}')

    # Return this boolean.
    return is_hint_recursive_state

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_hint_sane_recursable(
    hint_recursable: Hint,
    hint_nonrecursable: Hint,
    hint_parent_sane: Optional[HintSane],
) -> HintSane:
    '''
    **Sanified type hint metadata** (i.e., :class:`.HintSane` object) safely
    encapsulating both the passed **recursable type hint** (i.e., type hint
    implicitly supporting recursion like, say, a :pep:`695`-compliant type
    alias) and the passed metadata encapsulating the previously sanified parent
    type hint of the passed type hint.

    This factory creates and returns metadata protecting this recursable type
    hint against infinite recursion. Notably, this factory adds this hint to the
    :attr:`HintSane.hint_recursable_to_depth` instance variable of this metadata
    implementing the recursion guard for this hint.

    Parameters
    ----------
    hint_recursable : Hint
        Recursable type hint to be added to the
        :attr:`.HintSane.hint_recursable_to_depth` frozen set of the returned metadata.
    hint_nonrecursable : Hint
        Non-recursable type hint to be encapsulated if this type hint has both
        recursable and non-recursable forms describing this type hint. The
        distinction is as follows:

        * The recursable form of this type hint (passed as the mandatory
          ``hint_recursable`` parameter) is the variant of this hint that will
          be subsequently passed to the :func:`.is_hint_recursive` tester to
          detect whether this hint has already been recursively visited or not.
        * The non-recursable form of this type hint (passed as this optional
          ``hint_nonrecursable`` parameter) is the variant of this hint that
          will be encapsulated as the :attr:`.HintSane.hint` instance variable
          of the returned metadata. This non-recursable form is typically the
          post-sanified hint produced by sanifying the recursable form of the
          pre-sanified hint passed as the ``hint_recursable`` parameter.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If this recursable type hint is a root type hint, :data:`None`.
        * Else, **sanified parent type hint metadata** (i.e., immutable and thus
          hashable object encapsulating *all* metadata previously returned by
          :mod:`beartype._check.convert.convmain` sanifiers after sanitizing
          the possibly PEP-noncompliant parent hint of this hint into a fully
          PEP-compliant parent hint).

    Returns
    -------
    HintSane
        Sanified metadata encapsulating both:

        * This recursable type hint.
        * This sanified parent type hint metadata.
    '''
    assert hint_nonrecursable is not hint_recursable, (
        f'Non-recursable hint {repr(hint_nonrecursable)} == '
        f'recursable hint {repr(hint_recursable)}.'
    )
    assert isinstance(hint_parent_sane, NoneTypeOr[HintSane]), (
        f'{repr(hint_parent_sane)} neither sanified hint metadata nor "None".')

    # Sanified metadata to be returned.
    hint_sane: HintSane = None  # type: ignore[assignment]

    # If this hint has *NO* parent, this is a root hint. In this case...
    if hint_parent_sane is None:
        # Recursion guard recording the recursable form of this root hint to
        # have now been visited exactly once.
        hint_recursable_to_depth = FrozenDict({hint_recursable: 1})

        # Metadata encapsulating this hint and recursion guard.
        hint_sane = HintSane(
            hint=hint_nonrecursable,
            hint_recursable_to_depth=hint_recursable_to_depth,
        )
    # Else, this hint has a parent. In this case...
    else:
        # Total number of times this hint is permitted to have been previously
        # visited during the search from the root hint to this hint, trivially
        # defined as the prior total number of times plus one.
        hint_recursable_depth = (
            hint_parent_sane.hint_recursable_to_depth.get(hint_recursable, 0) +
            1
        )

        # Recursion guard recording the recursable form of this root hint to
        # have been visited this number of times.
        hint_recursable_to_depth = FrozenDict({
            hint_recursable: hint_recursable_depth})

        # If the parent hint is also associated with a recursion guard...
        if hint_parent_sane.hint_recursable_to_depth:
            # Full recursion guard merging the guard associated this parent hint
            # with the guard containing only this child hint, efficiently
            # defined as...
            #
            # Note that the order of operands in this "|" operation is
            # *SIGNIFICANT*. The guard protecting this hint takes precedence
            # over the guard protecting all transitive parent hints of this hint
            # and is thus intentionally passed as the second "|" operand.
            hint_recursable_to_depth = (
                # The guard protecting all transitive parent hints of this hint
                # with...
                hint_parent_sane.hint_recursable_to_depth |  # type: ignore[operator]
                # The guard protecting this hint.
                hint_recursable_to_depth
            )
        # Else, the parent hint is associated with *NO* such guard.

        # Metadata encapsulating this hint and recursion guard, while
        # "cascading" any other metadata associated with this parent hint (e.g.,
        # type variable lookup table) down onto this child hint as well.
        hint_sane = hint_parent_sane.permute_sane(
            hint=hint_nonrecursable,
            hint_recursable_to_depth=hint_recursable_to_depth,
        )

    # Return this underlying type hint.
    return hint_sane
